from collections import namedtuple
import logging
import os
import re

from configobj import ConfigObj
from validate import Validator, ValidateError

log = logging.getLogger(__name__)

VALID_NAME_RE = re.compile(r'^\w*\Z')
VALID_NET_ID_RE = re.compile(r'^0x[0-9a-fA-F]{4}$')


def _env_override():
    """Try to find SNAP_NETWORK_PATH env var"""

    try:
        return os.environ['SNAP_NETWORK_PATH']
    except KeyError:
        return None


def user_network_path():
    """Check for the environment var override or use the default"""
    return _env_override() or os.path.expanduser(os.path.join('~', '.snap', 'networks.cfg'))


SCNetwork = namedtuple('SCNetwork', ['channel', 'net_id', 'encryption_type', 'encryption_key', 'feature_bits',
                                     'default_group', 'default_ttl', 'default_delay',
                                     'default_cb_retries', 'default_cb_timeout'])


class Network(object):
    """Parameters defining a SNAP Network

    Networks define 'How do I talk to a SNAP network?'
    """

    _spec = [s.strip() for s in """
    channel = integer(min=0, max=64, default=4)
    net_id = string(min=0, max=6, default='0x1c2c')
    encryption_type = option('None', 'Basic', 'AES-128', '0', '1', '2', default='None')
    encryption_key = string(min=0, max=16, default='')
    feature_bits = string(min=0, max=6, default='0x0100')
    default_group = integer(min=0, max=65535, default=1)
    default_ttl = integer(min=0, max=255, default=5)
    default_delay = integer(min=0, max=255, default=50)
    default_cb_retries = integer(min=0, max=255, default=3)
    default_cb_timeout = integer(min=0, max=255, default=2)
    """.split('\n')]

    def __init__(self, name='default'):
        self.name = name
        self.channel = 4
        self.net_id = '0x1c2c'
        self.encryption_type = 'None'
        self.encryption_key = ''
        self.feature_bits = '0x0100'
        self.default_group = 1
        self.default_ttl = 5
        self.default_delay = 50
        self.default_cb_retries = 3
        self.default_cb_timeout = 2

    def __repr__(self):
        return "%s(%s: c=%d,n=%s,e=%s,f=%s, mcast_defaults=(%d, %d, %d), cb_defaults=(%d, %d))" % (self.__class__.__name__,
                                                                                                   self.name, self.channel,
                                                                                                   self.net_id,
                                                                                                   self.encryption_type,
                                                                                                   self.feature_bits,
                                                                                                   self.default_group,
                                                                                                   self.default_ttl,
                                                                                                   self.default_delay,
                                                                                                   self.default_cb_retries,
                                                                                                   self.default_cb_timeout)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def default():
        return Network()

    @staticmethod
    def list():
        user_cfg = ConfigObj(user_network_path())
        return user_cfg.sections

    @staticmethod
    def delete(name):
        user_cfg = ConfigObj(user_network_path())
        user_cfg.pop(name)
        user_cfg.write()

    @classmethod
    def load(cls, name):
        network = cls(name)
        log.debug("Reading Network: %s", name)
        v = Validator()
        cfg = ConfigObj(ConfigObj(user_network_path())[name], configspec=cls._spec)
        if cfg.validate(v):
            vars(network).update(**cfg)

        return network

    def validate(self):  # noqa  # too complex
        """Validate the network against the spec."""
        if not self.name:
            raise ValueError("Network must have a name")

        name = self.name.strip()
        if not VALID_NAME_RE.match(name):
            raise ValueError("Network Name must be a valid identifier")

        v = Validator()
        known_params = []
        for line in self._spec:
            if not line.strip():
                continue
            key, spec = [l.strip() for l in line.split('=', 1)]
            known_params.append(key)

            value = getattr(self, key)
            if value is None:
                raise ValueError("Invalid Network Parameter %s = %r" % (
                    key,
                    value,
                ))

            try:
                v.check(spec, value)
            except ValidateError as e:
                raise ValueError("Invalid Network Parameter %s = %r: %s" % (
                    key,
                    value,
                    e,
                ))

        # Any extraneous params?
        for k in vars(self):
            log.debug("checking param %r", k)
            if k not in known_params and k not in ('name'):
                raise ValueError("Extraneous Network Parameter %s", k)

        # Extra checks for network id and feature flags
        if not VALID_NET_ID_RE.match(self.net_id):
            raise ValueError("Invalid Network Parameter net_id = %r: Network ID must be 0x####" % self.net_id)

        if not VALID_NET_ID_RE.match(self.feature_bits):
            raise ValueError("Invalid Network Parameter feature_bits = %r: Feature Bits must be 0x####" % self.feature_bits)

    def save(self):
        """Save the network"""
        self.validate()

        user_cfg = ConfigObj(user_network_path())

        name = self.name.strip()
        user_cfg[name] = vars(self)
        # Don't write the name as a value in the section
        del user_cfg[name]['name']

        user_cfg.write()

    def as_sc_params(self):
        """Converts Network to SNAP Connect-ready parameters"""

        crypto_type = 0
        if self.encryption_type.lower() in ("aes-128", '1'):
            crypto_type = 1
        elif self.encryption_type.lower() in ("basic", '2'):
            crypto_type = 2

        return SCNetwork(self.channel, int(self.net_id, 16),
                         crypto_type, self.encryption_key,
                         int(self.feature_bits, 16),
                         self.default_group, self.default_ttl, self.default_delay,
                         self.default_cb_retries, self.default_cb_timeout)
