from collections import namedtuple
import binascii
import logging
import os
import re

from configobj import ConfigObj
from validate import Validator, ValidateError

log = logging.getLogger(__name__)

VALID_NAME_RE = re.compile(r'^\w*\Z')
VALID_COM_DEVICE_RE = re.compile(r'^COM([0-9]+)$')  # Serial on Windows
VALID_USB_DEVICE_RE = re.compile(r'^USB([0-9]+)$')  # SS100 - USB
VALID_SS_DEVICE_RE = re.compile(r'^SnapStick([0-9]+)$')  # SS200 - USB
VALID_RS232_DEVICE_RE = re.compile(r'^/dev/(snap|tty\.usbserial|ttyUSB|ttyO|ttyS|tty\.SLAB_USBtoUART).*$')  # Serial on Posix
VALID_ADDRESS_RE = re.compile(r'^[0-9A-Fa-f]{6}$')


def _env_override():
    """Try to find SNAP_PROFILE_PATH env var"""

    try:
        return os.environ['SNAP_PROFILE_PATH']
    except KeyError:
        return None


def user_profile_path():
    """Check for the environment var override or use the default"""
    return _env_override() or os.path.expanduser(os.path.join('~', '.snap', 'profiles.cfg'))


class Profile(object):
    """Parameters defining a SNAP Profile.

    Profiles define 'How do I connect to a SNAP network?'
    """

    def __init__(self, name="default"):
        self.name = name

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name)

    def __eq__(self, other):
        return (isinstance(other, self.__class__) and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def list():
        user_cfg = ConfigObj(user_profile_path())
        return user_cfg.sections

    @staticmethod
    def delete(name):
        user_cfg = ConfigObj(user_profile_path())
        user_cfg.pop(name)
        user_cfg.write()

    @classmethod
    def default(cls):
        return cls()

    @staticmethod
    def load(name):
        log.debug("Reading Profile: %s", name)

        # So...this is awkward. Need to read enough of the profile
        # to find out what kind it is, then switch on that

        raw_cfg = ConfigObj(user_profile_path())[name]

        log.debug("Raw Config: %r", raw_cfg)

        # sigh...
        if raw_cfg['kind'] == 'serial':
            cls = SerialProfile
        elif raw_cfg['kind'] == 'tcp':
            cls = TcpProfile
        elif raw_cfg['kind'] == 'mqtt':
            cls = MqttProfile

        log.debug("Profile Class: %r", cls)

        profile = cls(name)
        cfg = ConfigObj(raw_cfg, configspec=cls._spec, write_empty_values=True)

        log.debug("Config: %r", cfg)

        if cfg.validate(Validator()):
            vars(profile).update(**cfg)
        else:
            raise Exception("Validation failed!")

        return profile

    def validate(self):
        """Validate the Profile against the spec"""

        if not self.name:
            raise ValueError("Profile must have a name")

        name = self.name.strip()
        if not VALID_NAME_RE.match(name):
            raise ValueError("Profile Name must be a valid identifier")

        v = Validator()
        known_params = []
        for line in self._spec:
            if not line.strip():
                continue
            key, spec = [l.strip() for l in line.split('=', 1)]
            known_params.append(key)
            value = getattr(self, key)
            if value is None:
                raise ValueError("Invalid Profile Parameter %s = %r" % (
                    key,
                    value,
                ))
            try:
                v.check(spec, value)
            except ValidateError as e:
                raise ValueError("Invalid Profile Parameter %s = %r: %s" % (
                    key,
                    value,
                    e,
                ))

        # Any extraneous params?
        for k in vars(self):
            log.debug("checking param %r", k)
            if k not in known_params and k not in ('name'):
                raise ValueError("Extraneous Profile Parameter %s", k)

    def save(self):
        self.validate()

        user_cfg = ConfigObj(user_profile_path(), write_empty_values=True)
        name = self.name.strip()
        if not name:
            raise ValueError("Profile must have a name")

        user_cfg[name] = vars(self)
        # Don't write the name as a value in the section
        del user_cfg[name]['name']

        user_cfg.write()


SCSerialParams = namedtuple('SCSerialParams', ['type', 'port', 'address'])
SCTcpParams = namedtuple('SCTcpParams', ['host', 'port', 'ssl', 'username', 'password', 'address'])


class SerialProfile(Profile):
    """Serial Profile"""

    _spec = [s.strip() for s in """
    kind = option('serial')
    device = string(default='/dev/snap1')
    address = string(min=0, max=6, default='')
    network = string(default='default')
    """.split('\n')]

    def __init__(self, name='default'):
        super(SerialProfile, self).__init__(name)

        self.kind = "serial"
        self.device = "/dev/snap1"
        self.address = ""
        self.network = "default"

    def validate(self):
        """Validate the Profile against the spec"""

        super(SerialProfile, self).validate()

        if not (self.network == "" or VALID_NAME_RE.match(self.network)):
            raise ValueError("Invalid Profile Network = %r: Must be a valid identifier" % self.network)

        if not (self.address == "" or VALID_ADDRESS_RE.match(self.address)):
            raise ValueError("Invalid Profile Address = %r" % self.address)

        # Additional checks for Serial Profiles
        if not (VALID_COM_DEVICE_RE.match(self.device) or
                VALID_USB_DEVICE_RE.match(self.device) or
                VALID_SS_DEVICE_RE.match(self.device) or
                VALID_RS232_DEVICE_RE.match(self.device)):
            raise ValueError("Invalid Serial Profile Device = %r" % self.device)

    def _match_device(self):
        """Figure out what the SNAPconnect serial type and port are from our device string"""
        # Turn our device string into what SC expects
        usb_match = VALID_USB_DEVICE_RE.match(self.device)
        if usb_match:
            # USB SNAPStick 100, aka old paddleboard.
            # Why aren't you in VCP mode???
            serial_type = 2
            # USB0 -> port 0
            serial_port = int(usb_match.groups()[0])

            return serial_type, serial_port

        ss_match = VALID_SS_DEVICE_RE.match(self.device)
        if ss_match:
            serial_type = 5
            # SnapStick0 -> port 0
            serial_port = int(ss_match.groups()[0])

            return serial_type, serial_port

        com_match = VALID_COM_DEVICE_RE.match(self.device)
        if com_match:
            serial_type = 1
            # COM1 -> port 0
            serial_port = int(com_match.groups()[0]) - 1

            return serial_type, serial_port

        dev_match = VALID_RS232_DEVICE_RE.match(self.device)
        if dev_match:
            serial_type = 1
            serial_port = self.device

            return serial_type, serial_port

        raise ValueError("Invalid Serial Profile Device: %r" % self.device)

    def as_sc_params(self):
        """Translate from Profile to SC Params"""
        serial_type, serial_port = self._match_device()
        return SCSerialParams(serial_type, serial_port, binascii.unhexlify(self.address) if self.address else None)


class TcpProfile(Profile):
    """TCP Profile"""

    _spec = [s.strip() for s in """
    kind = option('tcp')
    host = string(min=1, default='127.0.0.1')
    port = integer(min=0, max=65535, default=48625)
    ssl = boolean(default=False)
    username = string(default='public')
    password = string(default='public')
    address = string(min=0, max=6, default='')
    network = string(default='default')
    """.split('\n')]

    def __init__(self, name='default'):
        super(TcpProfile, self).__init__(name)

        self.kind = "tcp"
        self.host = "127.0.0.1"
        self.port = 48625
        self.ssl = False
        self.username = "public"
        self.password = "public"
        self.address = ""
        self.network = "default"

    def validate(self):
        """Validate the Profile against the spec"""

        super(TcpProfile, self).validate()

        if not (self.address == "" or VALID_ADDRESS_RE.match(self.address)):
            raise ValueError("Invalid Profile Address = %r" % self.address)

        if not (self.network == "" or VALID_NAME_RE.match(self.network)):
            raise ValueError("Invalid Profile Network = %r: Must be a valid identifier" % self.network)

    def as_sc_params(self):
        return SCTcpParams(self.host, self.port, self.ssl, self.username, self.password,
                           binascii.unhexlify(self.address) if self.address else None)


class MqttProfile(Profile):
    """MQTT Profile"""

    _spec = [s.strip() for s in """
    kind = option('mqtt')
    host = string(min=1, default='127.0.0.1')
    port = integer(min=0, max=65535, default=1883)
    address = string(min=6, max=6, default='000080')
    ca_certs = string(default='')
    username = string(default='')
    password = string(default='')
    network = string(default='default')
    """.split('\n')]

    def __init__(self, name='default'):
        super(MqttProfile, self).__init__(name)
        self.kind = "mqtt"
        self.host = "127.0.0.1"
        self.port = 1883
        self.username = ''
        self.password = ''
        self.address = '000080'
        self.ca_certs = ''
        self.network = 'default'

    def validate(self):
        """Validate the Profile against the spec"""

        super(MqttProfile, self).validate()

        if not (VALID_ADDRESS_RE.match(self.address)):
            raise ValueError("Invalid Profile Address = %r" % self.address)

        if not (self.network == "" or VALID_NAME_RE.match(self.network)):
            raise ValueError("Invalid Profile Network = %r: Must be a valid identifier" % self.network)
